# -*- coding: UTF-8 -*-

import sys
import os
import marshal
import traceback
import json
from datetime import datetime
from qpid.messaging import Connection, Message, ConnectError
from qmf.console import Session
from common import logger, config, aif_objects
from common import statuses, rest_state
from common.statuses import upload as st
from upload.inserter import _Inserter, InserterError

class Inserter(_Inserter):
    checkedQueues = []
    def __init__(self, node):
        _Inserter.__init__(self, node)
        self.connection = None
        self.session = None
        self.sender = None
        self.existQueue = True
        
    def start(self):
        try:
            broker = '%s/%s@%s:%s' % (self.node.inserter.queue.username, self.node.inserter.queue.password, self.node.inserter.queue.host, str(self.node.inserter.queue.port))
            self.connection = Connection(broker)
            if not self.connection.opened():
                self.connection.open()
            self.session = self.connection.session(transactional=True)
            queueName = self.node.inserter.queue.queue
            if not queueName:
                queueName = 'artix.loaddict.%s' % self.node.id
            fileSize, fileCount, maxSize, maxCount = self.getQueueConfigs()
            self.existQueue = True
            if not queueName in Inserter.checkedQueues:
                self.existQueue = self.checkQueueParams(broker, queueName, maxSize, fileSize, fileCount, maxCount)
                Inserter.checkedQueues.append(queueName)
            self.sender = self.session.sender("%s; {create: always, node: { type: queue, durable: True, x-declare: {arguments: {'qpid.file_size':%s, 'qpid.file_count':%s, 'qpid.max_size':%s, 'qpid.max_count':%s}}} }" % (queueName, fileSize, fileCount, maxSize, maxCount), durable=True)
            prop = {}
            prop['type'] = 'control'
            prop['control'] = 'begin_tran'
            if logger.Logger().location == 'cs':
                prop['dictBackOfficeId'] = self.node.currentConverter.dict_state_id
            else:
                prop['dictCashLoadStateId'] = self.node.currentConverter.dict_state_id
                prop['cashCode'] = self.node.id
            prop['statusServer'] = self.node.currentConverter.statusServer if hasattr(self.node.currentConverter, 'statusServer') else 'localhost:8080'    
            self.log.info(u'Отправляем сообщение - начало транзакции')
            self.startMessageId = hex(hash(datetime.now()))
            prop['startMessageId'] = self.startMessageId
            self.sender.send(Message(content=None, id=self.startMessageId, properties=prop, durable=True))
            statuses.setUploadStatus(st.TRANSACTION_START)
            self.errorObjectsCount = 0
            return True
        except ConnectError, e:
            self.log.error(u'Ошибка: %s' % e)
            raise InserterError(u'Ошибка соединения с сервисом qpid, возможно он не запущен')
        except Exception, e:
            statuses.setUploadStatus(st.ERROR_TRANSACTION_START)
            self.log.error(u'Ошибка: %s' % e.message)
            self.log.error(traceback.format_exc())
        return False
    
    def add_package(self, listObjects):
        self.errorObjectsCount += aif_objects.checkObjects(listObjects, self.log)
        self.objectsCount += len(listObjects)
        self.log.debug(u'Отправляем пакет объектов справочника в очередь')
        statuses.setUploadStatus(st.ADD_PACKAGE)
        #self.log.debug('listObjects = %s' % listObjects)
        prop = {'type': 'datagram', 'serialType': 'marshal'}
        body = marshal.dumps(listObjects)
        self.sender.send(Message(content=body, id=hex(hash(datetime.now())), properties=prop, durable=True))
    
    def stop(self):
        if self.objectsCount == 0:
            self.log.warning(u'Ни одной записи не отправлено в очередь.')
            return self.rollback()
        prop = {'type': 'control', 'control':'end_tran'}
        self.log.info(u'Отправляем сообщение - конец транзакции')
        self.sender.send(Message(content=None, id=hex(hash(datetime.now())), properties=prop, durable=True))
        try:
            self.session.commit()
            if logger.Logger().location == 'cs':
                if not self.existQueue:
            	       rest_state.rereadQueuesInNES(self.log,self.node.nesServer)
                rest_state.startLoadNES(self.log,self.node.id,self.node.nesServer)
        except Exception, e:
            statuses.setUploadStatus(st.ERROR_TRANSACTION_END)
            self.log.error(u'Ошибка: %s' % (e.message))
            try:
                self.log.info(u'Отменяем запись в очередь')
                self.session.rollback()
                self.log.info(u'Закрываем connection')
                self.connection.close()
            except Exception, e:
                self.log.error(e.message)
            return False
        self.connection.close()
        statuses.setUploadStatus(st.TRANSACTION_END)
        self.log.info(u'Загрузка прошла успешно. Разблокируем загрузку в очередь для магазина - %s' % self.node.id)
        self.saveProgress()
        return True

    def invalidObjectsCount(self):
        return self.errorObjectsCount

    def rollback(self):
        self.log.info(u'Отменяем загрузку справочников в очередь для магазина - %s' % self.node.id)
        try:
            self.session.rollback()
        except:
            pass
        try:
            if self.connection and self.connection.opened():
                self.connection.close()
        except:
            pass
        return True
    
    def getQueueConfigs(self):
        fileSize = 128
        fileCount = 8
        maxSize = 104857600
        maxCount = 3000
        self.log.debug("sys.path = %s" % sys.path[0])
        confFileName = sys.path[0] + "/../config/%s/exchangers.ini" % (logger.Logger().location)
        if not config.read(confFileName):
            self.log.warning(u"Файл настроек '%s' не найден. Настройки очереди устанавливаются по-умолчанию..." % confFileName)
        else:
            fileSize = config.getInt('QUEUE', 'fileSize', fileSize)
            fileCount = config.getInt('QUEUE','fileCount', fileCount)
            maxSize = config.getInt('QUEUE','maxSize', maxSize)
            maxCount = config.getInt('QUEUE','maxCount', maxCount)
        return fileSize, fileCount, maxSize, maxCount
    
    def checkQueueParams(self, target, queueName, max_size, file_size, file_count, max_count):
        existQueue = False
        if os.name == 'posix':
            sess = None
            try:
                sess = Session()
                broker = sess.addBroker(target=target)
                queues = sess.getObjects(_class="queue", _package="org.apache.qpid.broker")
                for q in queues:
                    if q.name == queueName:
                        existQueue = True
                        args = q.arguments
                        if args.get('qpid.max_size', None) != max_size or args.get('qpid.file_size', None) != file_size or \
                            args.get('qpid.file_count', None) != file_count or args.get('qpid.max_count', None) != max_count:
                            if q.msgDepth == 0:
                                # если настройки очереди изменились, то удаляем эту очередь, чтобы создать с новыми настройками
                                broker.getAmqpSession().queue_delete(queue=q.name, if_empty=True, if_unused=True)
                        break
            except Exception, e:
                self.log.warning(u'Ошибка удалении очереди %s. %s' % (queueName, e.message))
            finally:
                if sess:
                    sess.close()
        return existQueue

    def saveProgress(self):
        _Inserter.saveProgress(self)
